/*
 * Copyright (c) 2009, 2013
 *	Craig Leres
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code
 * distributions retain the above copyright notice and this paragraph
 * in its entirety, and (2) distributions including binary code include
 * the above copyright notice and this paragraph in its entirety in
 * the documentation or other materials provided with the distribution
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
#ifndef lint
static const char copyright[] =
    "@(#) Copyright (c) 2009, 2013\nCraig Leres\n";
static const char rcsid[] =
    "@(#) $Id: main.c 43 2013-11-14 23:58:32Z leres $ (XSE)";
#endif

#include <sys/types.h>

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "tws.h"
#include "version.h"

/*
 * The number of days in the year, accounting for leap years
 */
#define dysize(y) \
    (((y) % 4) ? 365 : (((y) % 100) ? 366 : (((y) % 400) ? 365 : 366)))


/* Globals */
int debug;
int utc;
const char *prog;
const char *fmt = NULL;

/* Forwards */
time_t dmktime(struct tws *);
char *fmtargv(int, char **);
int main(int, char **);
void usage(void) __attribute__((noreturn));
void zonehack(struct tws *);

/* Locals */
static int dmsize[] = {
	31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
};

int
main(int argc, char **argv)
{
	char *cp;
	int op;
	struct tws *tw;
	struct tm *tm;
	time_t clock;
	char tstr[256];

	if ((cp = strrchr(argv[0], '/')) != NULL)
		prog = cp + 1;
	else
		prog = argv[0];

	while ((op = getopt(argc, argv, "duf:")) != EOF)
		switch (op) {

		case 'd':
			++debug;
			break;

		case 'f':
			if (*optarg == '\0')
				fmt = NULL;
			else
				fmt = optarg;
			break;

		case 'u':
			++utc;
			break;

		default:
			usage();
		}

	if (argc == optind)
		usage();

	cp = fmtargv(argc - optind, argv + optind);
	tw = dparsetime(cp);
	if (tw == NULL) {
		fprintf(stderr, "%s: Can't parse \"%s\"\n", prog, cp);
		exit(1);
	}
	if (tw->tw_year == 0) {
		fprintf(stderr, "%s: Can't find year (\"%s\")\n", prog, cp);
		exit(1);
	}

	if (debug)
		fprintf(stderr, "%d/%02d/%02d %d:%02d:%02d (%d)\n",
		    tw->tw_year, 1 + tw->tw_mon, tw->tw_mday,
		    tw->tw_hour, tw->tw_min, tw->tw_sec, tw->tw_zone);

	/* Set the timezone to local if it wasn't explicitly set */
	if ((tw->tw_flags & TW_SZONE) == TW_SZNIL) {
		tzset();
		time(&clock);
		tm = localtime(&clock);
		if (tm == NULL) {
			fprintf(stderr, "%s: localtime(%lu) failed\n",
			    prog, (u_long)clock);
			exit(1);
		}
		tw->tw_zone = tm->tm_gmtoff / 60;
		tw->tw_flags |= TW_SZEXP;
	}

	/* Unix seconds */
	clock = dmktime(tw);

	if (fmt == NULL) {
		if (utc)
			clock += tw->tw_zone * -60;
		printf("%ld\n", (long)clock);
	} else {
		if (utc)
			tm = gmtime(&clock);
		else
			tm = localtime(&clock);
		(void)strftime(tstr, sizeof(tstr), fmt, tm);
		printf("%s\n", tstr);
	}
	exit(0);
}

void
adios(const char *what, const char *fmt, ...)
{
	va_list ap;

	fprintf(stderr, "%s: ", prog);
	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	if (what != NULL)
		fprintf(stderr, " %s", what);
	fputc('\n', stderr);

	exit(1);
}

char *
fmtargv(int ac, char **av)
{
        int i;
        static char buf[256];
        char *cp;
        size_t len, size;

        cp = buf;
        size = sizeof(buf);

        for (i = 0; i < ac; ++i) {
                if (i > 0 && size > 1) {
                        *cp++ = ' ';
                        --size;
                }
                len = snprintf(cp, size, "%s", av[i]);
                if ((int)len < 0)
                        abort();
                cp += len;
                size -= len;
        }
        *cp = '\0';
        return (buf);
}
/*
 * Convert nmh time structure for local "broken-down"
 * time to calendar time (clock value).  This routine
 * is based on the gtime() routine written by Steven Shafer
 * at CMU.  It was forwarded to MTR by Jay Lepreau at Utah-CS.
 */
time_t
dmktime(struct tws *tw)
{
	int i, sec, min, hour, mday, mon, year;
	time_t result;

	if (tw->tw_clock != 0)
		return tw->tw_clock;

	if ((sec = tw->tw_sec) < 0 || sec > 61 ||
	    (min = tw->tw_min) < 0 || min > 59 ||
	    (hour = tw->tw_hour) < 0 || hour > 23 ||
	    (mday = tw->tw_mday) < 1 || mday > 31 ||
	    (mon = tw->tw_mon + 1) < 1 || mon > 12)
		return (tw->tw_clock = (time_t) -1);

	year = tw->tw_year;

	result = 0;
	if (year < 1970)
	    year += 1900;

	if (year < 1970)
		year += 100;

	for (i = 1970; i < year; i++)
		result += dysize(i);
	if (dysize(year) == 366 && mon >= 3)
		++result;
	while (--mon)
		result += dmsize[mon - 1];
	result += mday - 1;
	result = 24 * result + hour;
	result = 60 * result + min;
	result = 60 * result + sec;
	result -= 60 * tw->tw_zone;
	if (tw->tw_flags & TW_DST)
		result -= 60 * 60;

	return (tw->tw_clock = result);
}

void
usage(void)
{

	fprintf(stderr, "%s version %s\n", prog, version);
	fprintf(stderr, "usage: %s [-du] [-f fmt] time+date\n", prog);
	exit(1);
}
